@@ -59,6 +59,16 @@ char *mgcp_debug_get_last_endpoint_name(void)
 	return debug_last_endpoint_name;
 }
 
+const struct value_string mgcp_connection_mode_strs[] = {
+	{ MGCP_CONN_NONE, "none" },
+	{ MGCP_CONN_RECV_SEND, "sendrecv" },
+	{ MGCP_CONN_SEND_ONLY, "sendonly" },
+	{ MGCP_CONN_RECV_ONLY, "recvonly" },
+	{ MGCP_CONN_CONFECHO, "confecho" },
+	{ MGCP_CONN_LOOPBACK, "loopback" },
+	{ 0, NULL }
+};
+
 /* A combination of LOGPENDP and LOGPTRUNK that automatically falls back to
  * LOGPTRUNK when the endp parameter is NULL */
 #define LOGPEPTR(endp, trunk, cat, level, fmt, args...) \
@@ -172,7 +182,7 @@ static int setup_rtp_processing(struct mgcp_endpoint *endp,
 	/* Find the "sister" connection */
 	llist_for_each_entry(_conn, &endp->conns, entry) {
 		if (_conn->id != conn->conn->id) {
-			conn_src = &_conn->u.rtp;
+			conn_src = mgcp_conn_get_conn_rtp(_conn);
 			break;
 		}
 	}
@@ -479,46 +489,6 @@ static struct msgb *handle_audit_endpoint(struct mgcp_request_data *rq)
 	return create_ok_response(rq->trunk, rq->endp, 200, "AUEP", rq->pdata->trans);
 }
 
-/* Try to find a free port by attempting to bind on it. Also handle the
- * counter that points on the next free port. Since we have a pointer
- * to the next free port, binding should in work on the first attempt in
- * general. In case of failure the next port is tried until the whole port
- * range is tried once. */
-static int allocate_port(struct mgcp_endpoint *endp, struct mgcp_conn_rtp *conn)
-{
-	int i;
-	struct mgcp_port_range *range;
-	unsigned int tries;
-
-	OSMO_ASSERT(conn);
-
-	range = &endp->trunk->cfg->net_ports;
-
-	pthread_mutex_lock(&range->lock);
-	/* attempt to find a port */
-	tries = (range->range_end - range->range_start) / 2;
-	for (i = 0; i < tries; ++i) {
-		int rc;
-
-		if (range->last_port >= range->range_end)
-			range->last_port = range->range_start;
-
-		rc = mgcp_bind_net_rtp_port(endp, range->last_port, conn);
-
-		range->last_port += 2;
-		if (rc == 0) {
-			pthread_mutex_unlock(&range->lock);
-			return 0;
-		}
-
-	}
-	pthread_mutex_unlock(&range->lock);
-	LOGPENDP(endp, DLMGCP, LOGL_ERROR,
-	     "Allocating a RTP/RTCP port failed %u times.\n",
-	     tries);
-	return -1;
-}
-
 /*! Helper function for check_local_cx_options() to get a pointer of the next
  *  lco option identifier
  *  \param[in] lco string
@@ -712,41 +682,24 @@ static int set_local_cx_options(void *ctx, struct mgcp_lco *lco,
 	return 0;
 }
 
-void mgcp_rtp_end_config(struct mgcp_endpoint *endp, int expect_ssrc_change,
-			 struct mgcp_rtp_end *rtp)
-{
-	struct mgcp_trunk *trunk = endp->trunk;
-
-	int patch_ssrc = expect_ssrc_change && trunk->force_constant_ssrc;
-
-	rtp->force_aligned_timing = trunk->force_aligned_timing;
-	rtp->force_constant_ssrc = patch_ssrc ? 1 : 0;
-	rtp->rfc5993_hr_convert = trunk->rfc5993_hr_convert;
-
-	LOGPENDP(endp, DLMGCP, LOGL_DEBUG,
-		 "Configuring RTP endpoint: local port %d%s%s\n",
-		 osmo_sockaddr_port(&rtp->addr.u.sa),
-		 rtp->force_aligned_timing ? ", force constant timing" : "",
-		 rtp->force_constant_ssrc ? ", force constant ssrc" : "");
-}
-
 uint32_t mgcp_rtp_packet_duration(const struct mgcp_endpoint *endp,
 				  const struct mgcp_rtp_end *rtp)
 {
 	int f = 0;
+	struct mgcp_rtp_codec *codec = rtp->cset.codec;
 
 	/* Get the number of frames per channel and packet */
 	if (rtp->frames_per_packet)
 		f = rtp->frames_per_packet;
-	else if (rtp->packet_duration_ms && rtp->codec->frame_duration_num) {
-		int den = 1000 * rtp->codec->frame_duration_num;
-		f = (rtp->packet_duration_ms * rtp->codec->frame_duration_den +
+	else if (rtp->packet_duration_ms && codec->frame_duration_num) {
+		int den = 1000 * codec->frame_duration_num;
+		f = (rtp->packet_duration_ms * codec->frame_duration_den +
 		     den / 2)
 		    / den;
 	}
 
-	return rtp->codec->rate * f * rtp->codec->frame_duration_num /
-	    rtp->codec->frame_duration_den;
+	return codec->rate * f * codec->frame_duration_num /
+	    codec->frame_duration_den;
 }
 
 /*! Initializes osmux socket if not yet initialized. Parses Osmux CID from MGCP line.
@@ -774,6 +727,7 @@ static int handle_codec_info(struct mgcp_conn_rtp *conn,
 	struct mgcp_endpoint *endp = rq->endp;
 	struct mgcp_conn *conn_dst;
 	struct mgcp_conn_rtp *conn_dst_rtp;
+	struct mgcp_rtp_codecset *cset = &conn->end.cset;
 
 	int rc;
 	char *cmd;
@@ -787,7 +741,7 @@ static int handle_codec_info(struct mgcp_conn_rtp *conn,
 	if (have_sdp) {
 		/* If we have SDP, we ignore the local connection options and
 		 * use only the SDP information. */
-		mgcp_codec_reset_all(conn);
+		mgcp_codecset_reset(cset);
 		rc = mgcp_parse_sdp_data(endp, conn, rq->pdata);
 		if (rc != 0) {
 			LOGPCONN(conn->conn, DLMGCP,  LOGL_ERROR,
@@ -799,19 +753,19 @@ static int handle_codec_info(struct mgcp_conn_rtp *conn,
 	} else if (endp->local_options.codec) {
 		/* When no SDP is available, we use the codec information from
 		 * the local connection options (if present) */
-		mgcp_codec_reset_all(conn);
-		rc = mgcp_codec_add(conn, PTYPE_UNDEFINED, endp->local_options.codec, NULL);
+		mgcp_codecset_reset(cset);
+		rc = mgcp_codecset_add_codec(cset, PTYPE_UNDEFINED, endp->local_options.codec, NULL);
 		if (rc != 0)
 			goto error;
 	}
 
 	/* Make sure we always set a sane default codec */
-	if (conn->end.codecs_assigned == 0) {
+	if (cset->codecs_assigned == 0) {
 		/* When SDP and/or LCO did not supply any codec information,
 		 * than it makes sense to pick a sane default: (payload-type 0,
 		 * PCMU), see also: OS#2658 */
-		mgcp_codec_reset_all(conn);
-		rc = mgcp_codec_add(conn, 0, NULL, NULL);
+		mgcp_codecset_reset(cset);
+		rc = mgcp_codecset_add_codec(cset, 0, NULL, NULL);
 		if (rc != 0)
 			goto error;
 	}
@@ -819,12 +773,12 @@ static int handle_codec_info(struct mgcp_conn_rtp *conn,
 	/* Try to find an destination RTP connection that we can include in the codec decision. */
 	conn_dst = mgcp_find_dst_conn(conn->conn);
 	if (conn_dst && conn_dst->type == MGCP_CONN_TYPE_RTP)
-		conn_dst_rtp = &conn_dst->u.rtp;
+		conn_dst_rtp = mgcp_conn_get_conn_rtp(conn_dst);
 	else
 		conn_dst_rtp = NULL;
 
 	/* Make codec decision */
-	if (mgcp_codec_decide(conn, conn_dst_rtp) != 0)
+	if (mgcp_codecset_decide(&conn->end.cset, conn_dst_rtp ? &conn_dst_rtp->end.cset : NULL) != 0)
 		goto error;
 
 	return 0;
@@ -870,11 +824,11 @@ static struct msgb *handle_create_con(struct mgcp_request_data *rq)
 	int error_code = 400;
 	const char *local_options = NULL;
 	const char *callid = NULL;
-	const char *mode = NULL;
+	enum mgcp_connection_mode mode = MGCP_CONN_NONE;
 	char *line;
 	int have_sdp = 0, remote_osmux_cid = -2;
-	struct mgcp_conn_rtp *conn = NULL;
-	struct mgcp_conn *_conn = NULL;
+	struct mgcp_conn *conn = NULL;
+	struct mgcp_conn_rtp *conn_rtp = NULL;
 	char conn_name[512];
 	int rc;
 
@@ -922,7 +876,7 @@ static struct msgb *handle_create_con(struct mgcp_request_data *rq)
 			return create_err_response(rq->trunk, NULL, 523, "CRCX", pdata->trans);
 			break;
 		case 'M':
-			mode = (const char *)line + 3;
+			mode = mgcp_parse_conn_mode((const char *)line + 3);
 			break;
 		case 'X':
 			if (strncasecmp("Osmux: ", line + 2, strlen("Osmux: ")) == 0) {
@@ -959,15 +913,15 @@ mgcp_header_done:
 		return create_err_response(endp, endp, 516, "CRCX", pdata->trans);
 	}
 
-	if (!mode) {
+	if (mode == MGCP_CONN_NONE) {
 		LOGPENDP(endp, DLMGCP, LOGL_ERROR,
-			 "CRCX: insufficient parameters, missing mode\n");
+			 "CRCX: insufficient parameters, invalid mode\n");
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_INVALID_MODE));
 		return create_err_response(endp, endp, 517, "CRCX", pdata->trans);
 	}
 
 	/* Check if we are able to accept the creation of another connection */
-	if (endp->type->max_conns > 0 && llist_count(&endp->conns) >= endp->type->max_conns) {
+	if (mgcp_endp_is_full(endp)) {
 		LOGPENDP(endp, DLMGCP, LOGL_ERROR,
 			"CRCX: endpoint full, max. %i connections allowed!\n",
 			endp->type->max_conns);
@@ -975,7 +929,8 @@ mgcp_header_done:
 			/* There is no more room for a connection, make some
 			 * room by blindly tossing the oldest of the two two
 			 * connections */
-			mgcp_conn_free_oldest(endp);
+			mgcp_endp_free_conn_oldest(endp);
+			OSMO_ASSERT(!mgcp_endp_is_full(endp));
 		} else {
 			/* There is no more room for a connection, leave
 			 * everything as it is and return with an error */
@@ -1014,36 +969,35 @@ mgcp_header_done:
 	}
 
 	snprintf(conn_name, sizeof(conn_name), "%s", callid);
-	_conn = mgcp_conn_alloc(trunk->endpoints, endp, MGCP_CONN_TYPE_RTP, conn_name);
-	if (!_conn) {
+	conn = mgcp_conn_alloc(trunk->endpoints, endp, MGCP_CONN_TYPE_RTP, conn_name);
+	if (!conn) {
 		LOGPENDP(endp, DLMGCP, LOGL_ERROR,
 			 "CRCX: unable to allocate RTP connection\n");
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_ALLOC_CONN));
 		goto error2;
-
 	}
 
-	conn = mgcp_conn_get_rtp(endp, _conn->id);
-	OSMO_ASSERT(conn);
-
-	if (mgcp_parse_conn_mode(mode, endp, conn->conn) != 0) {
+	if (mgcp_conn_set_mode(conn, mode) < 0) {
 		error_code = 517;
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_INVALID_MODE));
 		goto error2;
 	}
 
+	conn_rtp = mgcp_conn_get_conn_rtp(conn);
+	OSMO_ASSERT(conn_rtp);
+
 	/* If X-Osmux (remote CID) was received (-1 is wilcard), alloc next avail CID as local CID */
 	if (remote_osmux_cid >= -1) {
-		if (osmux_init_conn(conn) < 0) {
+		if (osmux_init_conn(conn_rtp) < 0) {
 			rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_NO_OSMUX));
 			goto error2;
 		}
 		if (remote_osmux_cid >= 0) {
-			conn->osmux.remote_cid_present = true;
-			conn->osmux.remote_cid = remote_osmux_cid;
+			conn_rtp->osmux.remote_cid_present = true;
+			conn_rtp->osmux.remote_cid = remote_osmux_cid;
 		}
 	} else if (endp->trunk->cfg->osmux.usage == OSMUX_USAGE_ONLY) {
-		LOGPCONN(_conn, DLMGCP, LOGL_ERROR,
+		LOGPCONN(conn, DLMGCP, LOGL_ERROR,
 			 "CRCX: osmux only and no osmux offered\n");
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_NO_OSMUX));
 		goto error2;
@@ -1054,7 +1008,7 @@ mgcp_header_done:
 		rc = set_local_cx_options(trunk->endpoints,
 					  &endp->local_options, local_options);
 		if (rc != 0) {
-			LOGPCONN(_conn, DLMGCP, LOGL_ERROR,
+			LOGPCONN(conn, DLMGCP, LOGL_ERROR,
 				 "CRCX: invalid local connection options!\n");
 			error_code = rc;
 			rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_INVALID_CONN_OPTIONS));
@@ -1063,8 +1017,8 @@ mgcp_header_done:
 	}
 
 	/* Handle codec information and decide for a suitable codec */
-	rc = handle_codec_info(conn, rq, have_sdp, true);
-	mgcp_codec_summary(conn);
+	rc = handle_codec_info(conn_rtp, rq, have_sdp, true);
+	mgcp_codecset_summary(&conn_rtp->end.cset, mgcp_conn_dump(conn));
 	if (rc) {
 		error_code = rc;
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_CODEC_NEGOTIATION));
@@ -1072,60 +1026,54 @@ mgcp_header_done:
 	}
 	/* Upgrade the conn type RTP_DEFAULT->RTP_IUUP if needed based on requested codec: */
 	/* TODO: "codec" probably needs to be moved from endp to conn */
-	if (conn->type == MGCP_RTP_DEFAULT && strcmp(conn->end.codec->subtype_name, "VND.3GPP.IUFP") == 0) {
-		rc = mgcp_conn_iuup_init(conn);
+	if (conn_rtp->type == MGCP_RTP_DEFAULT &&
+	    strcmp(conn_rtp->end.cset.codec->subtype_name, "VND.3GPP.IUFP") == 0) {
+		rc = mgcp_conn_iuup_init(conn_rtp);
 	}
 
-	if (pdata->cfg->force_ptime) {
-		conn->end.packet_duration_ms = pdata->cfg->force_ptime;
-		conn->end.force_output_ptime = 1;
-	}
-
-	mgcp_rtp_end_config(endp, 0, &conn->end);
-
 	/* Find a local address for conn based on policy and initial SDP remote
 	   information, then find a free port for it */
-	if (mgcp_get_local_addr(conn->end.local_addr, conn) < 0) {
+	if (mgcp_get_local_addr(conn_rtp->end.local_addr, conn_rtp) < 0) {
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_BIND_PORT));
 		goto error2;
 	}
-	if (allocate_port(endp, conn) != 0) {
+	if (mgcp_trunk_allocate_conn_rtp_ports(trunk, conn_rtp) != 0) {
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_BIND_PORT));
 		goto error2;
 	}
 
-	if (setup_rtp_processing(endp, conn) != 0) {
-		LOGPCONN(_conn, DLMGCP, LOGL_ERROR,
+	if (setup_rtp_processing(endp, conn_rtp) != 0) {
+		LOGPCONN(conn, DLMGCP, LOGL_ERROR,
 			 "CRCX: could not start RTP processing!\n");
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_START_RTP));
 		goto error2;
 	}
 
 	/* Notify Osmux conn that CRCX was received */
-	if (mgcp_conn_rtp_is_osmux(conn)) {
-		if (conn_osmux_event_rx_crcx_mdcx(conn) < 0) {
-			LOGPCONN(conn->conn, DLMGCP, LOGL_ERROR, "CRCX: Osmux handling failed!\n");
+	if (mgcp_conn_rtp_is_osmux(conn_rtp)) {
+		if (conn_osmux_event_rx_crcx_mdcx(conn_rtp) < 0) {
+			LOGPCONN(conn, DLMGCP, LOGL_ERROR, "CRCX: Osmux handling failed!\n");
 			goto error2;
 		}
 	}
 
-	LOGPCONN(conn->conn, DLMGCP, LOGL_DEBUG,
-		 "CRCX: Creating connection: port: %u\n", conn->end.local_port);
+	LOGPCONN(conn, DLMGCP, LOGL_DEBUG,
+		 "CRCX: Creating connection: port: %u\n", conn_rtp->end.local_port);
 
 	/* Send dummy packet, see also comments in mgcp_keepalive_timer_cb() */
 	OSMO_ASSERT(trunk->keepalive_interval >= MGCP_KEEPALIVE_ONCE);
-	if (conn->conn->mode & MGCP_CONN_RECV_ONLY &&
+	if (conn->mode & MGCP_CONN_RECV_ONLY &&
 	    trunk->keepalive_interval != MGCP_KEEPALIVE_NEVER)
-		send_dummy(endp, conn);
+		send_dummy(endp, conn_rtp);
 
-	LOGPCONN(_conn, DLMGCP, LOGL_NOTICE,
+	LOGPCONN(conn, DLMGCP, LOGL_NOTICE,
 		 "CRCX: connection successfully created\n");
 	rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_SUCCESS));
 	mgcp_endp_update(endp);
 
 	/* NOTE: Only in the virtual trunk we allow dynamic endpoint names */
 	bool add_epname = rq->wildcarded && trunk->trunk_type == MGCP_TRUNK_VIRTUAL;
-	return create_response_with_sdp(endp, conn, "CRCX", pdata->trans, add_epname, true);
+	return create_response_with_sdp(endp, conn_rtp, "CRCX", pdata->trans, add_epname, true);
 error2:
 	mgcp_endp_release(endp);
 	LOGPENDP(endp, DLMGCP, LOGL_NOTICE,
@@ -1142,13 +1090,13 @@ static struct msgb *handle_modify_con(struct mgcp_request_data *rq)
 	struct rate_ctr_group *rate_ctrs;
 	char new_local_addr[INET6_ADDRSTRLEN];
 	int error_code = 500;
-	int silent = 0;
 	int have_sdp = 0;
 	char *line;
 	const char *local_options = NULL;
-	const char *mode = NULL;
-	struct mgcp_conn_rtp *conn = NULL;
-        const char *conn_id = NULL;
+	enum mgcp_connection_mode mode = MGCP_CONN_NONE;
+	struct mgcp_conn *conn = NULL;
+	struct mgcp_conn_rtp *conn_rtp = NULL;
+	const char *conn_id = NULL;
 	int remote_osmux_cid = -2;
 	int rc;
 
@@ -1175,7 +1123,7 @@ static struct msgb *handle_modify_con(struct mgcp_request_data *rq)
 		LOGPENDP(endp, DLMGCP, LOGL_ERROR, "MDCX: selected endpoint not available!\n");
 		return create_err_response(rq->trunk, NULL, 501, "MDCX", pdata->trans);
 	}
-	if (llist_count(&endp->conns) <= 0) {
+	if (mgcp_endp_num_conns(endp) <= 0) {
 		LOGPENDP(endp, DLMGCP, LOGL_ERROR,
 			 "MDCX: endpoint is not holding a connection.\n");
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_MDCX_FAIL_NO_CONN));
@@ -1205,10 +1153,7 @@ static struct msgb *handle_modify_con(struct mgcp_request_data *rq)
 			local_options = (const char *)line + 3;
 			break;
 		case 'M':
-			mode = (const char *)line + 3;
-			break;
-		case 'Z':
-			silent = strcasecmp("noanswer", line + 3) == 0;
+			mode = mgcp_parse_conn_mode((const char *)line + 3);
 			break;
 		case 'X':
 			if (strncasecmp("Osmux: ", line + 2, strlen("Osmux: ")) == 0) {
@@ -1242,29 +1187,30 @@ mgcp_header_done:
 		return create_err_response(endp, endp, 515, "MDCX", pdata->trans);
 	}
 
-	conn = mgcp_conn_get_rtp(endp, conn_id);
+	conn = mgcp_endp_get_conn(endp, conn_id);
 	if (!conn) {
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_MDCX_FAIL_CONN_NOT_FOUND));
 		return create_err_response(endp, endp, 400, "MDCX", pdata->trans);
 	}
 
-	mgcp_conn_watchdog_kick(conn->conn);
+	mgcp_conn_watchdog_kick(conn);
 
-	if (mode) {
-		if (mgcp_parse_conn_mode(mode, endp, conn->conn) != 0) {
+	if (mode != MGCP_CONN_NONE) {
+		if (mgcp_conn_set_mode(conn, mode) < 0) {
 			rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_MDCX_FAIL_INVALID_MODE));
 			error_code = 517;
 			goto error3;
 		}
-	} else
-		conn->conn->mode = conn->conn->mode_orig;
+	} else {
+		conn->mode = conn->mode_orig;
+	}
 
 	/* Set local connection options, if present */
 	if (local_options) {
 		rc = set_local_cx_options(trunk->endpoints,
 					  &endp->local_options, local_options);
 		if (rc != 0) {
-			LOGPCONN(conn->conn, DLMGCP, LOGL_ERROR,
+			LOGPCONN(conn, DLMGCP, LOGL_ERROR,
 				 "MDCX: invalid local connection options!\n");
 			error_code = rc;
 			rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_MDCX_FAIL_INVALID_CONN_OPTIONS));
@@ -1272,38 +1218,42 @@ mgcp_header_done:
 		}
 	}
 
+	conn_rtp = mgcp_conn_get_conn_rtp(conn);
+	OSMO_ASSERT(conn_rtp);
+
 	/* Handle codec information and decide for a suitable codec */
-	rc = handle_codec_info(conn, rq, have_sdp, false);
-	mgcp_codec_summary(conn);
+	rc = handle_codec_info(conn_rtp, rq, have_sdp, false);
+	mgcp_codecset_summary(&conn_rtp->end.cset, mgcp_conn_dump(conn));
 	if (rc) {
 		error_code = rc;
 		goto error3;
 	}
 	/* Upgrade the conn type RTP_DEFAULT->RTP_IUUP if needed based on requested codec: */
 	/* TODO: "codec" probably needs to be moved from endp to conn */
-	if (conn->type == MGCP_RTP_DEFAULT && strcmp(conn->end.codec->subtype_name, "VND.3GPP.IUFP") == 0)
-		rc = mgcp_conn_iuup_init(conn);
+	if (conn_rtp->type == MGCP_RTP_DEFAULT &&
+	    strcmp(conn_rtp->end.cset.codec->subtype_name, "VND.3GPP.IUFP") == 0)
+		rc = mgcp_conn_iuup_init(conn_rtp);
 
-	if (mgcp_conn_rtp_is_osmux(conn)) {
-		OSMO_ASSERT(conn->osmux.local_cid_allocated);
+	if (mgcp_conn_rtp_is_osmux(conn_rtp)) {
+		OSMO_ASSERT(conn_rtp->osmux.local_cid_allocated);
 		if (remote_osmux_cid < -1) {
-			LOGPCONN(conn->conn, DLMGCP, LOGL_ERROR,
+			LOGPCONN(conn, DLMGCP, LOGL_ERROR,
 				 "MDCX: Failed to parse Osmux CID!\n");
 			goto error3;
 		} else if (remote_osmux_cid == -1) {
-			LOGPCONN(conn->conn, DLMGCP, LOGL_ERROR,
+			LOGPCONN(conn, DLMGCP, LOGL_ERROR,
 				 "MDCX: wilcard in MDCX is not supported!\n");
 			goto error3;
-		} else if (conn->osmux.remote_cid_present &&
-				remote_osmux_cid != conn->osmux.remote_cid) {
-				LOGPCONN(conn->conn, DLMGCP, LOGL_ERROR,
-					"MDCX: changing already allocated CID is not supported!\n");
-				goto error3;
+		} else if (conn_rtp->osmux.remote_cid_present &&
+			   remote_osmux_cid != conn_rtp->osmux.remote_cid) {
+			LOGPCONN(conn, DLMGCP, LOGL_ERROR,
+				"MDCX: changing already allocated CID is not supported!\n");
+			goto error3;
 		} else {
-			conn->osmux.remote_cid_present = true;
-			conn->osmux.remote_cid = remote_osmux_cid;
-			if (conn_osmux_event_rx_crcx_mdcx(conn) < 0) {
-				LOGPCONN(conn->conn, DLMGCP, LOGL_ERROR,
+			conn_rtp->osmux.remote_cid_present = true;
+			conn_rtp->osmux.remote_cid = remote_osmux_cid;
+			if (conn_osmux_event_rx_crcx_mdcx(conn_rtp) < 0) {
+				LOGPCONN(conn, DLMGCP, LOGL_ERROR,
 					"MDCX: Osmux handling failed!\n");
 				goto error3;
 			}
@@ -1314,50 +1264,42 @@ mgcp_header_done:
 	   to update our announced IP addr and re-bind our local end. This can
 	   happen for instance if MGW initially provided an IPv4 during CRCX
 	   ACK, and now MDCX tells us the remote has an IPv6 address. */
-	if (mgcp_get_local_addr(new_local_addr, conn) < 0) {
+	if (mgcp_get_local_addr(new_local_addr, conn_rtp) < 0) {
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_BIND_PORT));
 		goto error3;
 	}
-	if (strcmp(new_local_addr, conn->end.local_addr)) {
-		osmo_strlcpy(conn->end.local_addr, new_local_addr, sizeof(conn->end.local_addr));
-		mgcp_free_rtp_port(&conn->end);
-		if (allocate_port(endp, conn) != 0) {
+	if (strcmp(new_local_addr, conn_rtp->end.local_addr)) {
+		osmo_strlcpy(conn_rtp->end.local_addr, new_local_addr, sizeof(conn_rtp->end.local_addr));
+		mgcp_rtp_end_free_port(&conn_rtp->end);
+		if (mgcp_trunk_allocate_conn_rtp_ports(trunk, conn_rtp) != 0) {
 			rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_CRCX_FAIL_BIND_PORT));
 			goto error3;
 		}
 	}
 
-	if (setup_rtp_processing(endp, conn) != 0) {
+	if (setup_rtp_processing(endp, conn_rtp) != 0) {
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_MDCX_FAIL_START_RTP));
 		goto error3;
 	}
 
-	mgcp_rtp_end_config(endp, 1, &conn->end);
-
 	/* modify */
-	LOGPCONN(conn->conn, DLMGCP, LOGL_DEBUG,
-		 "MDCX: modified conn:%s\n", mgcp_conn_dump(conn->conn));
+	LOGPCONN(conn, DLMGCP, LOGL_DEBUG,
+		 "MDCX: modified conn:%s\n", mgcp_conn_dump(conn));
 
 	/* Send dummy packet, see also comments in mgcp_keepalive_timer_cb() */
 	OSMO_ASSERT(trunk->keepalive_interval >= MGCP_KEEPALIVE_ONCE);
-	if (conn->conn->mode & MGCP_CONN_RECV_ONLY &&
+	if (conn->mode & MGCP_CONN_RECV_ONLY &&
 	    trunk->keepalive_interval != MGCP_KEEPALIVE_NEVER)
-		send_dummy(endp, conn);
+		send_dummy(endp, conn_rtp);
 
 	rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_MDCX_SUCCESS));
-	if (silent)
-		goto out_silent;
 
-	LOGPCONN(conn->conn, DLMGCP, LOGL_NOTICE,
+	LOGPCONN(conn, DLMGCP, LOGL_NOTICE,
 		 "MDCX: connection successfully modified\n");
 	mgcp_endp_update(endp);
-	return create_response_with_sdp(endp, conn, "MDCX", pdata->trans, false, false);
+	return create_response_with_sdp(endp, conn_rtp, "MDCX", pdata->trans, false, false);
 error3:
 	return create_err_response(endp, endp, error_code, "MDCX", pdata->trans);
-
-out_silent:
-	LOGPENDP(endp, DLMGCP, LOGL_DEBUG, "MDCX: silent exit\n");
-	return NULL;
 }
 
 /* DLCX command handler, processes the received command */
@@ -1368,11 +1310,10 @@ static struct msgb *handle_delete_con(struct mgcp_request_data *rq)
 	struct mgcp_endpoint *endp = rq->endp;
 	struct rate_ctr_group *rate_ctrs;
 	int error_code = 400;
-	int silent = 0;
 	char *line;
 	char stats[1048];
 	const char *conn_id = NULL;
-	struct mgcp_conn_rtp *conn = NULL;
+	struct mgcp_conn *conn = NULL;
 	unsigned int i;
 
 	/* NOTE: In this handler we can not take it for granted that the endp
@@ -1408,7 +1349,7 @@ static struct msgb *handle_delete_con(struct mgcp_request_data *rq)
 	if (rq->wildcarded) {
 		int num_conns = 0;
 		for (i = 0; i < trunk->number_endpoints; i++) {
-			num_conns += llist_count(&trunk->endpoints[i]->conns);
+			num_conns += mgcp_endp_num_conns(trunk->endpoints[i]);
 			mgcp_endp_release(trunk->endpoints[i]);
 		}
 		rate_ctr_add(rate_ctr_group_get_ctr(rate_ctrs, MGCP_DLCX_SUCCESS), num_conns);
@@ -1452,9 +1393,6 @@ static struct msgb *handle_delete_con(struct mgcp_request_data *rq)
 				goto error3;
 			}
 			break;
-		case 'Z':
-			silent = strcasecmp("noanswer", line + 3) == 0;
-			break;
 		default:
 			LOGPEPTR(endp, trunk, DLMGCP, LOGL_NOTICE, "DLCX: Unhandled MGCP option: '%c'/%d\n",
 				 line[0], line[0]);
@@ -1473,7 +1411,7 @@ static struct msgb *handle_delete_con(struct mgcp_request_data *rq)
 	 * that we drop all connections on that specific endpoint at once.
 	 * (See also RFC3435 Section F.7) */
 	if (!conn_id) {
-		int num_conns = llist_count(&endp->conns);
+		int num_conns = mgcp_endp_num_conns(endp);
 		LOGPENDP(endp, DLMGCP, LOGL_NOTICE,
 			 "DLCX: missing ci (connectionIdentifier), will remove all connections (%d total) at once\n",
 			 num_conns);
@@ -1490,39 +1428,33 @@ static struct msgb *handle_delete_con(struct mgcp_request_data *rq)
 	}
 
 	/* Find the connection */
-	conn = mgcp_conn_get_rtp(endp, conn_id);
+	conn = mgcp_endp_get_conn(endp, conn_id);
 	if (!conn) {
 		rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_DLCX_FAIL_INVALID_CONNID));
 		goto error3;
 	}
 	/* save the statistics of the current connection */
-	mgcp_format_stats(stats, sizeof(stats), conn->conn);
+	mgcp_format_stats(stats, sizeof(stats), conn);
 
 	/* delete connection */
-	LOGPCONN(conn->conn, DLMGCP, LOGL_DEBUG, "DLCX: deleting conn:%s\n",
-		 mgcp_conn_dump(conn->conn));
-	mgcp_conn_free(endp, conn_id);
+	LOGPCONN(conn, DLMGCP, LOGL_DEBUG, "DLCX: deleting conn:%s\n",
+		 mgcp_conn_dump(conn));
+	mgcp_conn_free(conn);
 	LOGPENDP(endp, DLMGCP, LOGL_NOTICE,
 		 "DLCX: connection successfully deleted\n");
 
 	/* When all connections are closed, the endpoint will be released
 	 * in order to be ready to be used by another call. */
-	if (llist_count(&endp->conns) <= 0) {
+	if (mgcp_endp_num_conns(endp) <= 0) {
 		mgcp_endp_release(endp);
 		LOGPENDP(endp, DLMGCP, LOGL_DEBUG, "DLCX: endpoint released\n");
 	}
 
 	rate_ctr_inc(rate_ctr_group_get_ctr(rate_ctrs, MGCP_DLCX_SUCCESS));
-	if (silent)
-		goto out_silent;
 	return create_ok_resp_with_param(endp, endp, 250, "DLCX", pdata->trans, stats);
 
 error3:
 	return create_err_response(endp, endp, error_code, "DLCX", pdata->trans);
-
-out_silent:
-	LOGPENDP(endp, DLMGCP, LOGL_DEBUG, "DLCX: silent exit\n");
-	return NULL;
 }
 
 /* RSIP command handler, processes the received command */
@@ -1630,7 +1562,7 @@ static void mgcp_keepalive_timer_cb(void *_trunk)
 		llist_for_each_entry(conn, &endp->conns, entry) {
 			if (conn->type == MGCP_CONN_TYPE_RTP &&
 			    conn->mode == MGCP_CONN_RECV_ONLY)
-				send_dummy(endp, &conn->u.rtp);
+				send_dummy(endp, mgcp_conn_get_conn_rtp(conn));
 		}
 	}
 
